/* tcfilter.c: input filter for T-Code/TUT-Code using tclib */
/*
 * :
 *   GNU screenѤT-Code/TUT-Codeϥե륿Ǥ
 *   screencolonץץ:exec .!. tcfilterΤ褦˼¹Ԥޤ
 *
 * ץ:
 *   -o ưľ夫T-Code/TUT-Code⡼ɤ򥪥ˤޤ
 *   -k <ȥ륭> T-Code/TUT-Code⡼ɤΥȥ륭ꤷޤ
 *      ǥեȤΥȥ륭CTRL-JǤ
 *      ȥ륭λϡCTRLȤȤ߹碌򼨤^Ȥ
 *      \θʸɤ8ʿ3ǻꤷޤ
 *      CTRL-Jꤹ: -k ^j ⤷ -k \012
 *
 * ץ:
 *   tclib<http://www.tcp-ip.or.jp/~tagawa/archive/index.html>ȤäƤޤ
 *   TUT-CodeξTUTбѥåŬѤtclibȤäƤ
 *
 *   tcfilterϤstdinɤǡstdout˽Ϥޤ
 *   screentcfilterexec .!.ǵưȡtcfilterstdout
 *   screenΥץꥱץ˷ӤĤ뤿ᡢ
 *   tcfilterstdoutؤνϤץꥱץޤ
 *     ܡ --stdin--> tcfilter --stdout--> ץꥱץ
 *   (ܺ٤screeninfoexec commandʬ򻲾)
 *
 *   Ūˤtclib°sample.cƱǤʲΤ褦ʰ㤤ޤ
 *   - stdinraw⡼ɤˤơ꥿󥭡򲡤ʤƤϤ褦ѹ
 *   - ư˻ꤹ륪ץ(-o, -k)ɲá
 *
 * Х:
 *   - 򤼽ѴѴԽʸɽޤ
 *   - T-Code/TUT-Code⡼ɤ󤫥դɽ򤷤ʤΤǡ
 *     ξ֤狼ʤʤ뤳Ȥޤ
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include "tclib.h"

static void parse_args(int argc, char *argv[]);
static void usage(void);
static int tinit(void);
static void tterm(void);
static void onsignal(int sig);

static int toggle_key = '\012';	/* CTRL-J */
static int default_on = 0;	/* whether tc_mode_on() on start */
static struct termios otty, ntty;
static int ttyraw = 0;

int
main(int argc, char *argv[])
{
	tc_param_t	*param;
	char	result[MAXLINE];
	char	local_echo[MAXLINE];
	int	tc_mode;
	unsigned char uc;

	parse_args(argc, argv);	/* set option values */
	param = tc_init_lib(NULL, TC_USE_TSERV);
	if (param == NULL) {
		fprintf(stderr, "failed to tc_init_lib().\n");
		exit(1);
	}
	if (isatty(0)) {
		if (tinit() < 0) {
			fprintf(stderr, "failed to setup terminal.\n");
			exit(3);
		}
		setbuf(stdout, NULL);
	}
	tc_mode = 0;
	if (default_on) {
		tc_mode_on(param);
		tc_mode = 1;
	}

	/* main loop */
	while (read(0, &uc, 1) == 1) {
		if (uc == toggle_key) {
			tc_mode ^= 1;
			if (tc_mode)
				tc_mode_on(param);
			else
				tc_mode_off(param);
			continue;
		}
		if (tc_mode) {
			if (tc_char(uc, result, local_echo, param)) {
				fputs(result, stdout);
				/* fputs(local_echo, stderr); */
			}
		}
		else {
			fputc(uc, stdout);
		}
	}

	tterm();
	tc_end_lib(param);
        return 0;
}

static void
parse_args(int argc, char *argv[])
{
	int i;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			fprintf(stderr, "unknown parameter: %s\n", argv[i]);
			exit(2);
		}
		switch (argv[i][1]) {
		case '?':
			usage();
			exit(0);
			break;
		case 'o':
			default_on = 1;
			break;
		case 'k':
			if (++i >= argc) {
				fprintf(stderr, "toggle key not specified for -k\n");
				exit(2);
			}
			switch (argv[i][0]) {
			case '\\':
				if (sscanf(argv[i]+1, "%o", &toggle_key) < 1) {
					fprintf(stderr, "toggle key specification error: %s\n", argv[i]);
					exit(2);
				}
				break;
			case '^':
				toggle_key = argv[i][1];
				if (islower(toggle_key) && toggle_key < 0x80) {
					toggle_key = toupper(toggle_key);
					toggle_key -= '@';
				}
				else if (toggle_key >= '@' && toggle_key <= '_') {
					toggle_key -= '@';
				}
				else {
					fprintf(stderr, "illegal toggle key character('%c') after '^'\n", toggle_key);
					exit(2);
				}
				break;
			default:
				toggle_key = argv[i][1];
				break;
			}
			break;
		default:
			fprintf(stderr, "unknown option: %s\n", argv[i]);
			exit(2);
			break;
		}
	}
}

static void
usage(void)
{
	fprintf(stderr,
	"Usage: tcfilter [-o] [-k <toggle_key>]\n"
	"   -o: enable tc mode on start.\n"
	"   -k: specify tc mode toggle key as ^k or octal \\ooo.\n"
	"       (ex: -k ^j or -k \\012 for CTRL-J)\n"
	"   -?: display this help message\n");
}

/* put stdin into a raw mode */
/* W㡼ɡƥ󥹡ؾܲUNIXץߥ󥰡٥ץ11.10 */
static int
tinit(void)
{
	if (tcgetattr(0, &otty) < 0)
		return -1;	/* not terminal? */

	signal(SIGINT, onsignal);
	signal(SIGQUIT, onsignal);
	signal(SIGTERM, onsignal);
	signal(SIGHUP, onsignal);

	ntty = otty;
	ntty.c_lflag &= ~(ECHO|ICANON|IEXTEN|ISIG);
	ntty.c_iflag &= ~(BRKINT|ICRNL|INPCK|ISTRIP|IXON);
	ntty.c_cflag &= ~(CSIZE|PARENB);
	ntty.c_cflag |= CS8;
	ntty.c_oflag &= ~(OPOST);
	ntty.c_cc[VMIN] = 1;
	ntty.c_cc[VTIME] = 0;

	if (tcsetattr(0, TCSAFLUSH, &ntty) < 0) {
		tterm();
		return -1;
	}
	ttyraw = 1;
	return 0;
}

static void
tterm(void)
{
	if (ttyraw)
		tcsetattr(0, TCSAFLUSH, &otty);
	ttyraw = 0;
}

static void
onsignal(int sig)
{
	tterm();
	exit(0);
}
